/*********************************************************************************
 *
 * cfa-10003-test.c
 *
 * Test code for the CFA-10003 demo board with 4 of the Crystalfontz
 * CFAG4265A 1" LCD and one of the CFA-10004 carrier boards.
 *
 * The CFAG4265A is based on a Sitronix ST7565P controller which has
 * the serial SPI interface enabled. This is a read-only interface to
 * the controller.
 *
 *********************************************************************************/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "cfa-10003.h"
#include "CFAO4265_logo.h"

extern uint8_t cfag4265_logo[];


void spi_init(void)
{
     /* Initialize SPI as master sck = fck/16, data clocked on rising
      * edge. The SPI clock may be set as high as 20MHz for the
      * ST7565P LCD controller on the CFAG4265.
      */
     SPCR = (1 << SPE) | (1 << MSTR) | (1 << SPR0);

}


void spi_master_send_byte(uint8_t data)
{
     /* Start data transmission
      */
     SPDR = data;

     /* Wait for transmission complete
      */
     while (!(SPSR & (1 << SPIF)));
}


void lcd_reset(void)
{
     /* Set the LCD_RST low for 10 uS and then high again and return
      * after 10 uS. According to the ST7565P datasheet only 1 uS is
      * needed at VDD = 3.3VDC but it won't hurt to be conservative.
      */
     PORTB &= ~(1 << LCD_RST);
     _delay_us(10);
     PORTB |= (1 << LCD_RST);
     _delay_us(10);
}


void lcd_cmd_send(uint8_t disp, uint8_t data)
{
     /* Set LCD1_CS low and LCD_A0 low to indicate command to LCD Send
      * command byte to LCD
      */
     LCD_CTRL &= ~((1 << disp) | (1 << LCD_A0));
     spi_master_send_byte(data);

     /* set LCD1_CS high and LCD_A0 high again
      */
     LCD_CTRL |= ((1 << disp) | (1 << LCD_A0));

}


void lcd_data_send(uint8_t disp, uint8_t data)
{
     /* Set LCD1_CS low and LCD_A0 high to indicate data to LCD Send
      * data byte to LCD
      */
     LCD_CTRL &= ~(1 << disp);
     LCD_CTRL |= (1 << LCD_A0);
     spi_master_send_byte(data);

     /* set LCD1_CS high and LCD_A0 high again
      */
     LCD_CTRL |= ((1 << disp) | (1 << LCD_A0));
}



void lcd_init(uint8_t disp)
{
     lcd_bias_set(disp, LCD_BIAS_1_7);
     lcd_adc_select(disp, NORMAL);
     lcd_common_output_mode_select(disp, NORMAL);
     lcd_vo_regulator_resistor_set(disp, LCD_RATIO_5_0);
     lcd_electronic_volume_set(disp, 32);
     lcd_booster_ratio_set(disp, LCD_BOOSTER_4X);
     lcd_power_controller_set(disp, BOOST_ON, VR_ON, VF_ON);
     lcd_start_line_set(disp, 0);
     lcd_clear(disp);
}



void lcd_display_on(uint8_t disp)
{
     lcd_cmd_send(disp, 0b10101111);
}

void lcd_display_off(uint8_t disp)
{
     lcd_cmd_send(disp, 0b10101110);
}

void lcd_start_line_set(uint8_t disp, uint8_t line)
{
     lcd_cmd_send(disp, 0b01000000 | (line & 0b111111));
}

void lcd_page_address_set(uint8_t disp, uint8_t page)
{
     lcd_cmd_send(disp, 0b10110000 | (page & 0b1111));
}

void lcd_column_address_set(uint8_t disp, uint8_t column)
{
     lcd_cmd_send(disp, 0b00010000 | (column >> 4));
     lcd_cmd_send(disp, 0b00000000 | (column & 0b1111));
}

void lcd_adc_select(uint8_t disp, uint8_t direction)
{
     lcd_cmd_send(disp, 0b10100000 | (direction & 0b1));
}

void lcd_display_reverse(uint8_t disp, uint8_t reverse)
{
     lcd_cmd_send(disp, 0b101000110 | (reverse & 0b1));
}

void lcd_all_points(uint8_t disp, uint8_t on)
{
     lcd_cmd_send(disp, 0b10100100 | (on & 0b1));
}

void lcd_bias_set(uint8_t disp, uint8_t bias)
{
     lcd_cmd_send(disp, 0b10100010 | (bias & 0b1));
}

void lcd_cmd_reset(uint8_t disp)
{
     lcd_cmd_send(disp, 0b11100010);
}

void lcd_common_output_mode_select(uint8_t disp, uint8_t reverse)
{
     lcd_cmd_send(disp, 0b11000000 | ((reverse & 0b1) << 3));
}

void lcd_power_controller_set(uint8_t disp, uint8_t boost, uint8_t vreg, uint8_t vfollower)
{
     lcd_cmd_send(disp, 0b00101000 | ((boost & 0b1) << 2)
                  | ((vreg & 0b1) << 1)
                  | (vfollower & 0b1));
}

void lcd_electronic_volume_set(uint8_t disp, uint8_t contrast)
{
     lcd_cmd_send(disp, 0b10000001);
     lcd_cmd_send(disp, contrast & 0b00111111);
}

void lcd_vo_regulator_resistor_set(uint8_t disp, uint8_t ratio)
{
     lcd_cmd_send(disp, 0b00100000 | (ratio & 0b111));
}

void lcd_booster_ratio_set(uint8_t disp, uint8_t ratio)
{
     lcd_cmd_send(disp, 0b11111000);
     lcd_cmd_send(disp, ratio & 0b11);
}

void lcd_clear(uint8_t disp)
{
     uint8_t page, col;

     for (page = 0; page < 9; ++page)
     {
          lcd_page_address_set(disp, page);
          lcd_column_address_set(disp, 0);
          for (col = 0; col < 42; ++col)
          {
               lcd_data_send(disp, 0);
          }
     }
}

void lcd_fill(uint8_t disp, uint8_t data, uint16_t delay_ms_col, uint16_t delay_ms_page)
{
     uint8_t page, col;

     for (page = 0; page < 9; ++page)
     {
          lcd_page_address_set(disp, page);
          lcd_column_address_set(disp, 0);
          for (col = 0; col < 42; ++col)
          {
               lcd_data_send(disp, data);
               _delay_ms(delay_ms_col);

          }
          _delay_ms(delay_ms_page);
     }
}

void lcd_backlight_set(uint8_t brightness)
{
     dac_cmd_send(DAC_REG_WRITE_UPDATE, brightness);
}


void dac_init(void)
{
     /* Initialize the LTC2630-8 DAC. Set its reference to VCC
      * (+3.3VDC)
      */
     dac_cmd_send(DAC_POWER_DOWN, NULL_DATA);
     dac_cmd_send(DAC_VCC_REF, NULL_DATA);

}


void dac_cmd_send(dac_cmd_t cmd, uint8_t data)
{
     /* Set DAC_CS low. Send DAC command, data byte followed by don't
      * care byte to DAC
      */
     PORTA &= ~(1 << DAC_CS);
     spi_master_send_byte(cmd);
     spi_master_send_byte(data);
     spi_master_send_byte(NULL_DATA);

     /* Set DAC_CS high
      */
     PORTA |= (1 << DAC_CS);
}

void lcd_init_backlight_pwm_1(uint8_t brightness)
{
     /* Stop timer 0, disable the interrupt and reset the timer
      * count.
      */
     TCCR0A = 0;
     TIMSK0 = 0;
     TCNT0 = 0;

     /* Set the initial brightness PWM value
      * OCRA0 = 0, backlight full off
      * OCRA0 = 255, backlight full on
      */
     OCR0A = brightness;

     /* Using phase correct PWM mode
      * PWM Frequency = FOSC / (N * 510) where N = prescale factor
      *               = 8Mhz / (8 * 510) = 1.96KHz
      */
     TCCR0A = (1 << WGM00) | (1 << COM0A1);
     TCCR0B = (1 << CS01);

}

void lcd_set_backlight_1(uint8_t brightness)
{
     OCR0A = brightness;
}

void lcd_init_backlight_pwm_2(uint8_t brightness)
{

     /* Stop timer 1, disable the interrupt and reset the timer count.
      */
     TCCR1A = 0;
     TCCR1B = 0;
     TIMSK1 = 0;
     TCNT1 = 0;

     /* Set the initial brightness PWM value
      * OCRA1 = 0, backlight full off
      * OCRA1 = 255, backlight full on
      */
     OCR1A = brightness;

     /* Using phase correct PWM mode
      * PWM Frequency = FOSC / (2 * N * TOP) where N = prescale factor
      *               = 8Mhz / (2 * 8 * 255) = 1.96KHz
      */
     TCCR1A = (1 << WGM10) | (1 << COM1A1);
     TCCR1B = (1 << CS11);

}

void lcd_set_backlight_2(uint8_t brightness)
{
     OCR1A = brightness;
}

void lcd_init_backlight_pwm_3(uint8_t brightness)
{

     /* Stop timer 2, disable the interrupt and reset the timer count.
      */
     TCCR2A = 0;
     TCCR2B = 0;
     TIMSK2 = 0;
     TCNT2 = 0;

     /* Set the initial brightness PWM value
      * OCRA2 = 0, backlight full off
      * OCRA2 = 255, backlight full on
      */
     OCR2A = brightness;

     /* Using phase correct PWM mode
      * PWM Frequency = FOSC / (2 * N * TOP) where N = prescale factor
      *               = 8Mhz / (2 * 8 * 255) = 1.96KHz
      */
     TCCR2A = (1 << WGM20) | (1 << COM2A1);
     TCCR2B = (1 << CS21);

}

void lcd_set_backlight_3(uint8_t brightness)
{
     OCR2A = brightness;
}

void lcd_init_backlight_pwm_4(uint8_t brightness)
{

     /* Stop timer 3, disable the interrupt and reset the timer count.
      */
     TCCR3A = 0;
     TCCR3B = 0;
     TIMSK3 = 0;
     TCNT3 = 0;

     /* Set the initial brightness PWM value
      * OCRA3 = 0, backlight full off
      * OCRA3 = 255, backlight full on
      */
     OCR3A = brightness;

     /* Using phase correct PWM mode
      * PWM Frequency = FOSC / (2 * N * TOP) where N = prescale factor
      *               = 8Mhz / (2 * 8 * 255) = 1.96KHz
      */
     TCCR3A = (1 << WGM30) | (1 << COM3A1);
     TCCR3B = (1 << CS31);

}

void lcd_set_backlight_4(uint8_t brightness)
{
     OCR3A = brightness;
}



void gpio_init(void)
{
     /* Initialize PORT A
      *
      * bit  [7]   output, high     (LCD_A0)
      * bit  [6]   output, high     (LCD5_CS)
      * bit  [5]   output, high     (LCD3_CS1)
      * bit  [4]   output, high     (DAC_CS)
      * bit  [3]   output, high     (LCD4_CS)
      * bit  [2]   output, high     (LCD2_CS)
      * bit  [1]   output, high     (LCD1_CS)
      * bit  [0]   output, low      (NC)
      */
     DDRA  = 0b11111111;
     PORTA = 0b11111111;

     /* Initialize PORT B
      *
      * bit  [7]   output, high     (LCD1_PWM)
      * bit  [6]   output, high     (LCD_RST)
      * bit  [5]   output, low      (LCD2_PWM)
      * bit  [4]   output, low      (LCD4_PWM)
      * bit  [3]   input, no pullup (SPI MISO)
      * bit  [2]   output, low      (SPI MOSI)
      * bit  [1]   output, low      (SPI SCK)
      * bit  [0]   output, high     (SD_CS)
      */
     DDRB  = 0b11110111;
     PORTB = 0b11000001;

     /* Initialize PORT C - Not Used
      */
     DDRC  = 0b00000000;
     PORTC = 0b11111111;

     /* Initialize PORT D
      *
      * bit  [7]   input,  pullup   (SW4)
      * bit  [6]   input,  pullup   (SW3)
      * bit  [5]   input,  pullup   (SW2)
      * bit  [4]   input,  pullup   (SW1)
      * bit  [3]   output, low      (NC)
      * bit  [2]   output, low      (NC)
      * bit  [1]   output, low      (LED2)
      * bit  [0]   output, low      (LED1)
      */
     DDRD  = 0b00001111;
     PORTD = 0b11110000;

     DDRE  = 0b00001000;
     PORTE = 0b11110111;
}

void delay_100ms(uint8_t tenth_seconds)
{
     do
     {
          _delay_ms(100);
     } while (tenth_seconds-- != 0);
}


int main(void)
{
     uint8_t page, col, data1, data2;
     int16_t i = 0;
     uint8_t disp = 1;

     /* Initialize the hardware
      */
     gpio_init();
     spi_init();
     dac_init();
     lcd_reset();

     lcd_init(LCD1_CS);
     lcd_init(LCD2_CS);
     lcd_init(LCD3_CS);
     lcd_init(LCD4_CS);
     lcd_init(LCD5_CS);

     /* Turn on the display and its backlight
      */
     lcd_display_on(LCD1_CS);
     lcd_display_on(LCD2_CS);
     lcd_display_on(LCD3_CS);
     lcd_display_on(LCD4_CS);
     lcd_display_on(LCD5_CS);

     lcd_backlight_set(BACKLIGHT_FULL_ON);
     lcd_init_backlight_pwm_1(255);
     lcd_init_backlight_pwm_2(255);
     lcd_init_backlight_pwm_3(255);
     lcd_init_backlight_pwm_4(255);


     /* Loop through the demo forever
      */
     for (;;)
     {

          /* Display the Crystalfontz CFAO4265 bitmap image.
           *
           * The bitmap2lcd utility generated the data with the image
           * shifted up 7 pixels from the bottom of the LCD. This is
           * handled prior to display of each byte by shifting the
           * byte from the current page down 7 pixels and ORing it
           * with the byte from the next page that is shifted up 1
           * pixel.
           */
          for (i = 0, page = 0; page < 9; ++page)
          {
               lcd_page_address_set(LCD1_CS, page);
               lcd_page_address_set(LCD2_CS, page);
               lcd_page_address_set(LCD3_CS, page);
               lcd_page_address_set(LCD4_CS, page);
               lcd_page_address_set(LCD5_CS, page);

               lcd_column_address_set(LCD1_CS, 0);
               lcd_column_address_set(LCD2_CS, 0);
               lcd_column_address_set(LCD3_CS, 0);
               lcd_column_address_set(LCD4_CS, 0);
               lcd_column_address_set(LCD5_CS, 0);

               for (col = 0; col < 42; ++col)
               {
                    data1 = pgm_read_byte(&(cfag4265_logo[i]));
                    data2 = pgm_read_byte(&(cfag4265_logo[i+42]));

                    lcd_data_send(LCD1_CS, (data1 >> 7) | (data2 << 1));
                    lcd_data_send(LCD2_CS, (data1 >> 7) | (data2 << 1));
                    lcd_data_send(LCD3_CS, (data1 >> 7) | (data2 << 1));
                    lcd_data_send(LCD4_CS, (data1 >> 7) | (data2 << 1));
                    lcd_data_send(LCD5_CS, (data1 >> 7) | (data2 << 1));

                    ++i;
               }
          }

          /* Cycle the backlights
           */
          for (i = 255; i >= 0; --i)
          {

               lcd_backlight_set(i);
               lcd_set_backlight_1(i);
               lcd_set_backlight_2(i);
               lcd_set_backlight_3(i);
               lcd_set_backlight_4(i);
               delay_100ms(1);
          }
          delay_100ms(200);

          for (i = 0; i <= 255; ++i)
          {
               lcd_backlight_set(i);
               lcd_set_backlight_1(i);
               lcd_set_backlight_2(i);
               lcd_set_backlight_3(i);
               lcd_set_backlight_4(i);
               delay_100ms(1);
          }
          delay_100ms(200);


          lcd_clear(LCD1_CS);
          lcd_clear(LCD2_CS);
          lcd_clear(LCD3_CS);
          lcd_clear(LCD4_CS);
          lcd_clear(LCD5_CS);


          /* Display a fill pattern then clear
           */
          lcd_fill(LCD1_CS, 0x3c, 3, 1);
          lcd_fill(LCD2_CS, 0x3c, 3, 1);
          lcd_fill(LCD3_CS, 0x3c, 3, 1);
          lcd_fill(LCD4_CS, 0x3c, 3, 1);
          lcd_fill(LCD5_CS, 0x3c, 3, 1);

          lcd_fill(LCD1_CS, 0x00, 3, 1);
          lcd_fill(LCD2_CS, 0x00, 3, 1);
          lcd_fill(LCD3_CS, 0x00, 3, 1);
          lcd_fill(LCD4_CS, 0x00, 3, 1);
          lcd_fill(LCD5_CS, 0x00, 3, 1);

          /* Display a fill pattern then clear
           */
          lcd_fill(LCD1_CS, 0x03c, 0, 255);
          lcd_fill(LCD2_CS, 0x03c, 0, 255);
          lcd_fill(LCD3_CS, 0x03c, 0, 255);
          lcd_fill(LCD4_CS, 0x03c, 0, 255);
          lcd_fill(LCD5_CS, 0x03c, 0, 255);

          lcd_fill(LCD1_CS, 0x00, 0, 255);
          lcd_fill(LCD2_CS, 0x00, 0, 255);
          lcd_fill(LCD3_CS, 0x00, 0, 255);
          lcd_fill(LCD4_CS, 0x00, 0, 255);
          lcd_fill(LCD5_CS, 0x00, 0, 255);
     }
}
